// modified version: AR Pruss, 2008, opt key check

// modified version: performance patches
// daniel grobe sachs (dgsachs@nekito.net)

/* Main code for KeyCaps600 */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <NotifyMgr.h>
#include <Hs.h>
#include <Common/System/HsKeyTypes.h>
#include <Window.h>
#include "KeyCaps600Rsc.h"
#include "KeyCaps600.h"

#define ByteSwap16(n)   (((((UInt16) n) << 8) & 0xFF00) | \
                         ((((UInt16) n) >> 8) & 0x00FF))

static void RegisterForNotifications(Int8 priority);
static void DeregisterNotifications();
static Boolean isTreo600();
static void butlerNotify(SysNotifyParamType *np);
static void InitializePrefs( KeyCaps600Prefs *prefs );

static Int32 GetTextFieldLen( void );

// Protect the database? (Should be unnecessary)
//#define PROTECT_DB		

// Don't recognize hold for second/third taps of double-tap
#define SUPPRESS_HOLD

// Try to keep fast-paced double taps from being recognized as holds
#define WATCH_KEYUP

static Err HandleNotification(SysNotifyParamType *cmdPBP)
{
  EventPtr eventP;
  DWord pChr;
  UInt32 pMod;
  WChar optChr;

  UInt32 sHold,notHeld;

  UInt16 optModifiers, soundVol = 0;
  UInt32 ticks, count, tempDisable, syncing, popup;
  KeyCaps600Prefs prefs;
  Boolean repeated, held=0;
  UInt16 initDelayP, periodP, doubleTapDelayP;
  Boolean queueAheadP, capsLock, numLock, autoShift, optLock;
  UInt16 tempShift;
  UInt16 alarmVol, sysVol, gameVol;
  Int32 expectedLen, actualLen;
  UInt32 val;
  
  UInt16 keycodes[1];
  Boolean pressed[1];

  eventP = (EventPtr)(cmdPBP->notifyDetailsP);

  // If this is an event, make sure we care.
  if (cmdPBP->notifyType == sysNotifyEventDequeuedEvent) 
  {

    // Not keyDown? Don't care
    if( eventP->eType != ByteSwap16(keyDownEvent) )
      return 0;

#if 0
{ Char s[55]; Char f[16];
f[0]='%'; f[1]='x'; f[2]=' ';f[3]='%'; f[4]='x'; f[5]=' ';f[6]='%'; f[7]='x'; f[8] = 0;
StrPrintF(s, f, eventP->data.keyDown.chr, eventP->data.keyDown.keyCode, eventP->data.keyDown.modifiers);
WinDrawChars(s,StrLen(s),1,SysRandom(0)%150); }
#endif

    // Synthesized/graffiti? Don't care
    if( eventP->data.keyDown.keyCode == ByteSwap16(0) )
      return 0;

    // Not actually pressed? Don't care.
    keycodes[0] = HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr));

    if (HsKeysPressed(1, keycodes, pressed) == 0) 
      /* if this key isn't actually pressed, return */
      return 0;

    // Keys that don't have shift/alt can get skipped if we've already set
    // the repeat rate. Improves performance for games
    if( ((eventP->data.keyDown.modifiers & ByteSwap16(virtualKeyMask)) &&
	 (eventP->data.keyDown.chr != ByteSwap16(hsChrSymbol))) || 
	((eventP->data.keyDown.chr == ByteSwap16(chrBackspace)) || 
	 (eventP->data.keyDown.chr == ByteSwap16(chrFullStop)) ||
	 (eventP->data.keyDown.chr == ByteSwap16(chrCarriageReturn))) )
    {
	    val = 0;
	    FtrGet(KeyCaps600Creator, RepeatSetID, &val);
	    if( val )
	      return 0;
    }
  }

  /* do nothing with these events while syncing */
  if (cmdPBP->notifyType == sysNotifySyncStartEvent) {
    FtrSet(KeyCaps600Creator, HotsyncID, 1);
    return 0;
  }
  if (cmdPBP->notifyType == sysNotifySyncFinishEvent) {
    FtrSet(KeyCaps600Creator, HotsyncID, 0);
    return 0;
  }

  if (FtrGet(KeyCaps600Creator, HotsyncID, &syncing) == ftrErrNoSuchFeature) {
    syncing = 0;
  }
  if (syncing) return 0;
  
  // Check if we're deleting the application
  if( cmdPBP->notifyType == sysNotifyDeleteProtectedEvent )
  {
    if( ((SysNotifyDBInfoType *)cmdPBP->notifyDetailsP)->creator == KeyCaps600Creator )
    {
      DeregisterNotifications();
      cmdPBP->handled = true;
    }

    return 0;
  }


  /* get the prefs */
  InitializePrefs( &prefs );
  tempDisable = 0;

  // Repeat rate might not be set
  FtrSet(KeyCaps600Creator, RepeatSetID, false);

  if( ! (prefs.prefsMask & (HoldDownMask)) )
  {
    // Non-repeat functions are disabled, set the repeat rate
    KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
    periodP = prefs.scrollrepeat;
    initDelayP = prefs.scrolldelay;
    KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
    FtrSet(KeyCaps600Creator, RepeatSetID, true);
  }

  if( ! (prefs.prefsMask & (HoldDownMask | DoubleClickMask)) )
    return 0; // Nothing else to do - quit now

  if( prefs.prefsMask & DisableInPhone ) 
  {
    UInt16 cardId;
    UInt32 dbId;
    UInt32 val;
    UInt16 phone_cardid;
    UInt32 phone_dbid;

    SysCurAppDatabase( &cardId, &dbId );
    
    val = 0; FtrGet(KeyCaps600Creator, CardIdID, &val); phone_cardid = val;
    val = 0; FtrGet(KeyCaps600Creator, DBIdID  , &val); phone_dbid = val;

    if( (phone_cardid == cardId) && (phone_dbid == dbId) )
      tempDisable = 1;
  }

  if (cmdPBP->notifyType == sysNotifyEventDequeuedEvent) {
    // Are we in phone app?
    if( tempDisable )
      return 0; 

    ticks = 0; FtrGet(KeyCaps600Creator, TicksID, &ticks);
    pChr = -1; FtrGet(KeyCaps600Creator, ChrID, &pChr);
    pMod = 0; FtrGet(KeyCaps600Creator, ModID, &pMod);

    expectedLen = -1; FtrGet(KeyCaps600Creator, ExpectedLenID, &expectedLen);

#ifdef SUPPRESS_HOLD  // Do we suppress hold 
    sHold = 0; FtrGet(KeyCaps600Creator, SuppressHoldID, &sHold);

    if( ByteSwap16(eventP->data.keyDown.modifiers) & autoRepeatKeyMask ) 
    {
      if( sHold ) 
      { 
        // Eat the keystroke

        butlerNotify(cmdPBP);
        eventP->eType = ByteSwap16(nilEvent);
        return 0;
      }
      else
        held = true;
    }
#else
    held = ByteSwap16(eventP->data.keyDown.modifiers) & autoRepeatKeyMask;
#endif

#ifdef WATCH_KEYUP
    // is this a keyup?
    if( 0x140 == (ByteSwap16(eventP->data.keyDown.modifiers)) ) //& 0x140) )
    {
      // For the last key that went down?
      if( eventP->data.keyDown.chr == pChr )
        FtrSet(KeyCaps600Creator,NotHeldID, 1);
        
      // don't touch keyups
      return 0;
    }

    notHeld = 0; FtrGet(KeyCaps600Creator, NotHeldID, &notHeld);

    // if we saw a keyup, it clearly wasn't held
    if( notHeld )
    {
       held = 0;
       FtrSet(KeyCaps600Creator, NotHeldID, 0); // might be held now though
    }
#endif

    if( eventP->data.keyDown.chr == pChr )
      repeated = (TimGetTicks() <= ticks+(prefs.timeout));
    else
      repeated = 0;

    if( !isTreo600(&prefs) )
    {
      // On a T650, zero and symbol get different processing

      /* Eat repeating zeros if hold is active, because that's confusing */
      if( (eventP->data.keyDown.chr == ByteSwap16(chrDigitZero)) &&
                held &&
                prefs.prefsMask & HoldDownMask )
      {
	butlerNotify(cmdPBP);
        eventP->eType = ByteSwap16(nilEvent); 
        return 0;
      }

      /* Eat double-taps for zero if doubleclick is active */
      if( (eventP->data.keyDown.chr == ByteSwap16(chrDigitZero)) && 
		repeated &&
		eventP->data.keyDown.chr == pChr &&
		prefs.prefsMask & DoubleClickMask )
      {
        FtrSet(KeyCaps600Creator, TicksID, (DWord)TimGetTicks());  // reset clock

        butlerNotify(cmdPBP);
        eventP->eType = ByteSwap16(nilEvent);
        return 0;
      }
	
      if( eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol) )
      {
        // Set nav repeat rate
        KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
        periodP = prefs.scrollrepeat;
        initDelayP = prefs.scrolldelay;
        KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
        FtrSet(KeyCaps600Creator, RepeatSetID, true);

        return 0;
      } 
    } 

    
    if( isTreo600(&prefs) )
      soundVol = PrefGetPreference(prefSysSoundVolume);

    if (!repeated && isTreo600(&prefs) ) {
      /* restore sound volume */
      SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
      sysVol = soundVol;
      SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
    }
    
    if (FtrGet(KeyCaps600Creator, CountID, &count) == ftrErrNoSuchFeature) {
      count = 1;
    }

    if (FtrGet(KeyCaps600Creator, PopUpID, &popup) == ftrErrNoSuchFeature) {
      popup = 0;
    }

    if( ((eventP->data.keyDown.chr >= ByteSwap16(chrSmall_A)) && 
	 (eventP->data.keyDown.chr <= ByteSwap16(chrSmall_Z))) || 
        (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) )
    {
          FtrSet(KeyCaps600Creator, SuppressHoldID, false);
 	  keycodes[0] = HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr));
	  if (HsKeysPressed(1, keycodes, pressed) == 0) {
	    /* if this key isn't actually pressed, return */
	    return 0;
	  }
	
	  /* if we're capslocked or optionlocked, we do nothing */
	  HsGrfGetStateExt(&capsLock, &numLock, &optLock, &tempShift, &autoShift);
	  if (optLock || numLock || capsLock || ( tempShift /* & 4 ) || ( tempShift & 2 */ ) ) {
	    return 0;
	  }
#if 0
{ Char s[55]; Char f[16];
f[0]='%'; f[1]='x'; f[2]=' ';f[3]='%'; f[4]='x'; f[5]=' ';f[6]='%'; f[7]='x'; f[8] = 0;
StrPrintF(s, f, tempShift, autoShift, eventP->data.keyDown.modifiers);
WinDrawChars(s,StrLen(s),1,SysRandom(0)%150); }
#endif

	  /* if the key is the alt/0 key and we're temp-option, do nothing */
	  if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol) && (tempShift & hsGrfTempShiftOpt)) {
	    return 0;
	  }
	  
	  /* if a modifier key is down, we do nothing */
	  if ((ByteSwap16(eventP->data.keyDown.modifiers) &
	       (shiftKeyMask | commandKeyMask | optionKeyMask | controlKeyMask)) &&
	      (eventP->data.keyDown.chr != ByteSwap16(hsChrSymbol))) {
	    return 0;
	  }

	  /* set intial keyrate.. only if hold down does something */
	  if (prefs.prefsMask & HoldDownMask) {
	    KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	    initDelayP = prefs.delay;
	    KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	  }

	  // Save field length when key is initially pressed
	  if( (prefs.prefsMask & DetectAutocomplete) && 
	      !((prefs.prefsMask & DoubleClickMask) && repeated) &&
	      !((prefs.prefsMask & HoldDownMask) && held ))
          {
	    expectedLen = GetTextFieldLen();
	    if( expectedLen >= 0 ) expectedLen++;
	    FtrSet(KeyCaps600Creator,ExpectedLenID,expectedLen);
	  }

	  if (pChr == eventP->data.keyDown.chr) {

	    /* set slow subsequent keyrate.. only if hold down does something */
	    if (prefs.prefsMask & HoldDownMask) {
	      KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	      periodP = prefs.delay;
	      KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	    }
	  
	    if (repeated && !held  && (prefs.prefsMask & DoubleClickMask)) {
              FtrSet(KeyCaps600Creator, SuppressHoldID, true);
	      if (count == 1) {
		FtrSet(KeyCaps600Creator, CountID, 2);
		if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) {
		  if (prefs.prefsMask & DoubleClickOption) {
		    /* only backspace if popped up list - if winhandle is not the same as last time Alt pressed */
		    if (popup != (UInt32)WinGetActiveWindow()) {
		      /* Remove the popped up list */
		      eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		      eventP->data.keyDown.keyCode = 0;
		      eventP->data.keyDown.modifiers = 0;
		    } else {
		      /* Ignore the key */
                      butlerNotify(cmdPBP);
		      eventP->eType = ByteSwap16(nilEvent);
		    }
		    /* restore system sound volume */
		    SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		    sysVol = soundVol;
		    SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      

		    /* Insert the digit zero */
		    EvtEnqueueKey(chrDigitZero, 0, 0);
		  } else { /* DoubleClickCapital */
		    /* Ignore the key */
                    butlerNotify(cmdPBP);
		    eventP->eType = ByteSwap16(nilEvent);
		    /* restore system sound volume */
                    if( isTreo600(&prefs) )
                    { 
		      SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		      sysVol = soundVol;
		      SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
                    }
		    /* Don't insert a new character */
		  }
		} else {
		  if (prefs.prefsMask & DoubleClickCapital) {
		    /* Turn lower case into upper case */
		    optChr = ByteSwap16(eventP->data.keyDown.chr - ByteSwap16(0x20));
		    optModifiers = 0;
		  } else { /* DoubleClickOption */
		    /* special case to make contrast key work */
		    if (ByteSwap16(eventP->data.keyDown.chr) == chrSmall_P) {
                      if( prefs.prefsMask & DoubleClickRepeat ) { 
			optChr = 'P';
			optModifiers = 0;
                      } else {
#ifdef CONTRAST
		        optChr = vchrContrast;
		        optModifiers = commandKeyMask;
#else
			optChr = 'p';
			optModifiers = 0;
#endif
                      }
		    } else { /* Turn lower case into option */
		      HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)),
					    optionKeyMask, (UInt16*)&optChr, &optModifiers);
		    }
		  }
		  /* Remove the existing character */
		  eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		  eventP->data.keyDown.keyCode = 0;
		  eventP->data.keyDown.modifiers = 0;

		  if( prefs.prefsMask & DetectAutocomplete )
	          {
		    /* Eat autocomplete */
		    actualLen = GetTextFieldLen();
		    if( (expectedLen >= 0) && (actualLen >= 0) && (actualLen > expectedLen) )
		      EvtEnqueueKey(chrBackspace, 0, 0);
		  }

		  /* Insert the new character */
		  EvtEnqueueKey((WChar)optChr, 0, optModifiers);
		}
		if (prefs.prefsMask & DoubleClickRepeat) {
		  FtrSet(KeyCaps600Creator, TicksID, (DWord)TimGetTicks());
		} else if ((pChr != ByteSwap16(hsChrSymbol)) || (prefs.prefsMask & DoubleClickOption)) {
		  FtrSet(KeyCaps600Creator, ChrID, -1);
		  FtrSet(KeyCaps600Creator, ModID, 0);
		  FtrSet(KeyCaps600Creator, TicksID, -1);
		}
	      } else if ((count == 2) && (prefs.prefsMask & DoubleClickRepeat)) {
		FtrSet(KeyCaps600Creator, CountID, 3);
		if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) {
		  /* only backspace if popped up list - if winhandle is not the same as last time Alt pressed */
		  if (popup != (UInt32)WinGetActiveWindow()) {
		    /* Remove the popped up list */
		    eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		    eventP->data.keyDown.keyCode = 0;
		    eventP->data.keyDown.modifiers = 0;
		  } else {
		    /* Ignore the key */
		    butlerNotify(cmdPBP);
		    eventP->eType = ByteSwap16(nilEvent);
		  }
		  /* restore system sound volume */
		  SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		  sysVol = soundVol;
		  SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
		  if (prefs.prefsMask & DoubleClickCapital) {
		    /* Insert the digit zero */
		    EvtEnqueueKey(chrDigitZero, 0, 0);
		  }
		} else {
		  if (prefs.prefsMask & DoubleClickCapital) {
		    /* special case to make contrast key work */
		    if (ByteSwap16(eventP->data.keyDown.chr) == chrSmall_P) {
                      if( prefs.prefsMask & DoubleClickRepeat ) { 
			optChr = 'P';
			optModifiers = 0;
                      } else {
#ifdef CONTRAST
		        optChr = vchrContrast;
		        optModifiers = commandKeyMask; 
#else
			optChr = 'p';
			optModifiers = 0;
#endif
                      }
		    } else {
		      /* Turn lower case into option */
		      HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)),
					    optionKeyMask, (UInt16*)&optChr, &optModifiers);
		    }
		  } else { /* DoubleClickOption */
		    /* Turn lower case into upper case */
		    HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)), 
					  shiftKeyMask, (UInt16*)&optChr, &optModifiers);
		  }
		  /* Remove the existing character */
		  eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		  eventP->data.keyDown.keyCode = 0;
		  eventP->data.keyDown.modifiers = 0;

		  if( prefs.prefsMask & DetectAutocomplete )
	          {
		    /* Eat autocomplete */
		    actualLen = GetTextFieldLen();
		    if( (expectedLen >= 0) && (actualLen >= 0) && (actualLen > expectedLen) )
		      EvtEnqueueKey(chrBackspace, 0, 0);
		  }

		  /* Insert the new character */
		  EvtEnqueueKey((WChar)optChr, 0, optModifiers);
		}
		if ((pChr != ByteSwap16(hsChrSymbol)) || (prefs.prefsMask & DoubleClickCapital)) {
		  FtrSet(KeyCaps600Creator, ChrID, -1);
		  FtrSet(KeyCaps600Creator, ModID, 0);
		  FtrSet(KeyCaps600Creator, TicksID, -1);
		}
	      }

	    } else if (held && (prefs.prefsMask & HoldDownMask)) {

	      if (count == 1) {
		FtrSet(KeyCaps600Creator, CountID, 2);
		if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) {
		  /* only backspace if popped up list - if winhandle is not the same as last time Alt pressed */
		  if (popup != (UInt32)WinGetActiveWindow()) {
		    /* Remove the popped up list */
		    eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		    eventP->data.keyDown.keyCode = 0;
		    eventP->data.keyDown.modifiers = 0;
		  } else {
		    /* Ignore the key */
                    butlerNotify(cmdPBP);
		    eventP->eType = ByteSwap16(nilEvent);
		  }
		  /* restore system sound volume */
		  SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		  sysVol = soundVol;
		  SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
		  if (prefs.prefsMask & HoldDownOption) {
		    /* Insert the digit zero */
		    EvtEnqueueKey(chrDigitZero, 0, 0);
		  }
		} else {
		  if (prefs.prefsMask & HoldDownCapital) {
		    /* Turn lower case into upper case */
		    optChr = ByteSwap16(eventP->data.keyDown.chr - ByteSwap16(0x20));
		    optModifiers = 0;
		  } else { /* HoldDownOption */
		    /* special case to make contrast key work */
		    if (ByteSwap16(eventP->data.keyDown.chr) == chrSmall_P) {
                      if( prefs.prefsMask & HoldDownRepeat ) { 
                        optChr = 'P';
			optModifiers = 0;
                      } else {
#ifdef CONTRAST
		        optChr = vchrContrast;
		        optModifiers = commandKeyMask;
#else
			optChr = 'p';
			optModifiers = 0;
#endif
                      }
		    } else { /* Turn lower case into option */
		      HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)), 
					    optionKeyMask, (UInt16*)&optChr, &optModifiers);
		    }
		  }
		  /* Remove the existing character */
		  eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		  eventP->data.keyDown.keyCode = 0;
		  eventP->data.keyDown.modifiers = 0;

		  if( prefs.prefsMask & DetectAutocomplete )
	          {
		    /* Eat autocomplete */
		    actualLen = GetTextFieldLen();
		    if( (expectedLen >= 0) && (actualLen >= 0) && (actualLen > expectedLen) )
		      EvtEnqueueKey(chrBackspace, 0, 0);
		  }

		  /* Insert the new character */
		  EvtEnqueueKey((WChar)optChr, 0, optModifiers);
		}
	      } else if ((count == 2) && (prefs.prefsMask & HoldDownRepeat)) {
		FtrSet(KeyCaps600Creator, CountID, 3);
		if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) {
		  /* only backspace if popped up list - if winhandle is not the same as last time Alt pressed */
		  if (popup != (UInt32)WinGetActiveWindow()) {
		    /* Remove the popped up list */
		    eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		    eventP->data.keyDown.keyCode = 0;
		    eventP->data.keyDown.modifiers = 0;
		  } else {
		    /* Ignore the key */
                    butlerNotify(cmdPBP);
		    eventP->eType = ByteSwap16(nilEvent);
		  }
		  /* restore system sound volume */
		  SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		  sysVol = soundVol;
		  SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
		  if (prefs.prefsMask & HoldDownCapital) {
		    /* Insert the digit zero */
		    EvtEnqueueKey(chrDigitZero, 0, 0);
		  }
		} else {
		  if (prefs.prefsMask & HoldDownCapital) {
		    /* special case to make contrast key work */
		    if (ByteSwap16(eventP->data.keyDown.chr) == chrSmall_P) {
                      if( prefs.prefsMask & HoldDownRepeat ) { 
			optChr = 'P';
			optModifiers = 0;
                      } else {
#ifdef CONTRAST
		        optChr = vchrContrast;
		        optModifiers = commandKeyMask;
#else
			optChr = 'p';
			optModifiers = 0;
#endif
                      }
		    } else {
		      /* Turn lower case into option */
		      HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)), 
					    optionKeyMask, (UInt16*)&optChr, &optModifiers);		    }
		  } else { /* HoldDownOption */
		    /* Turn option into upper case */
		    HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)), 
					  shiftKeyMask, (UInt16*)&optChr, &optModifiers);
		  }
		  /* Remove the existing character */
		  eventP->data.keyDown.chr = ByteSwap16(chrBackspace);
		  eventP->data.keyDown.keyCode = 0;
		  eventP->data.keyDown.modifiers = 0;

		  if( prefs.prefsMask & DetectAutocomplete )
	          {
		    /* Eat autocomplete */
		    actualLen = GetTextFieldLen();
		    if( (expectedLen >= 0) && (actualLen >= 0) && (actualLen > expectedLen) )
		      EvtEnqueueKey(chrBackspace, 0, 0);
		  }

		  /* Insert the new character */
		  EvtEnqueueKey((WChar)optChr, 0, optModifiers);
		}
	      } else {
 		butlerNotify(cmdPBP);
		eventP->eType = ByteSwap16(nilEvent);
	      }
	    } else if (!repeated) {
	      FtrSet(KeyCaps600Creator, ChrID, eventP->data.keyDown.chr);
	      FtrSet(KeyCaps600Creator, ModID, eventP->data.keyDown.modifiers);
	      FtrSet(KeyCaps600Creator, CountID, 1);
	      FtrSet(KeyCaps600Creator, TicksID, TimGetTicks());
	      /* save window handle so we can see if the list pops up */
	      if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) {
		FtrSet(KeyCaps600Creator, PopUpID, (UInt32)WinGetActiveWindow());
		/* turn off system sounds */
		SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		sysVol = 0;
		SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);
	      } else {
		/* restore sound volume */
		if( isTreo600(&prefs) )
                { 
		  SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
		  sysVol = soundVol;
		  SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
                }
	      }
	    }
	  } else {

	    /* set fast subsequent keyrate.. only if hold down does something */
	    if (prefs.prefsMask & HoldDownMask) {
	      KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	      periodP = prefs.delay / 50;
	      KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	    }
	  
	    /* different char than previous */
	    FtrSet(KeyCaps600Creator, ChrID, eventP->data.keyDown.chr);
	    FtrSet(KeyCaps600Creator, ModID, eventP->data.keyDown.modifiers);
	    FtrSet(KeyCaps600Creator, CountID, (DWord)1);
	    FtrSet(KeyCaps600Creator, TicksID, TimGetTicks());
	    /* save window handle so we can see if the list pops up */
	    if (eventP->data.keyDown.chr == ByteSwap16(hsChrSymbol)) {
	      FtrSet(KeyCaps600Creator, PopUpID, (UInt32)WinGetActiveWindow());
	      /* turn off system sounds */
	      SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
	      sysVol = 0;
	      SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);
	    } else {
	      /* restore sound volume */
              if( isTreo600(&prefs) ) 
              { 
	        SndGetDefaultVolume(&alarmVol, &sysVol, &gameVol);
	        sysVol = soundVol;
	        SndSetDefaultVolume(&alarmVol, &sysVol, &gameVol);	      
              }
	    }
	  }

    } else { /* not a keyboard character, nor is it 'Alt' */
	  	  
	  /* identify the base keycode of this character */
	  HsKeyKeyCodeToChrCode(HsKeyChrCodeToKeyCode(ByteSwap16(eventP->data.keyDown.chr)),
				0, (UInt16*)&optChr, &optModifiers);

	  if (
	      /* check for a pseudo backspace */
	      ( (eventP->data.keyDown.chr == ByteSwap16(chrBackspace)) &&
		(eventP->data.keyDown.keyCode != ByteSwap16(chrBackspace)) ) ||

	      /* check to see if we're upper of previous */
	      ( (eventP->data.keyDown.chr == pChr - ByteSwap16(0x20)) &&
		( ( (prefs.prefsMask & (HoldDownCapital | DoubleClickCapital)) && (count == 2) ) ||
		  ( (prefs.prefsMask & (HoldDownOption | DoubleClickOption)) && (count == 3) ) ) ) ||
	      /* check to see if we're option of previous */
	      ( (ByteSwap16(optChr) == pChr) &&
		( ( (prefs.prefsMask & (HoldDownOption | DoubleClickOption)) && (count == 2) ) ||
		  ( (prefs.prefsMask & (HoldDownCapital | DoubleClickCapital)) && (count == 3) ) ) )
	      ) {
	    
	    /* set slow subsequent keyrate.. only if hold down does something */
	    if (prefs.prefsMask & HoldDownMask) {
	      KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	      periodP = prefs.delay;
	      KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	    }

	  } else {
	    /* set actual repeat rate */
	    if (prefs.prefsMask & HoldDownMask) {
	      KeyRates(0, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	      periodP = prefs.scrollrepeat;
	      initDelayP = prefs.scrolldelay;
	      KeyRates(1, &initDelayP, &periodP, &doubleTapDelayP, &queueAheadP);
	      FtrSet(KeyCaps600Creator, RepeatSetID, true);
	    }
	  
	    /* not keyboard key or backspace or hold down value previous key */
	    FtrSet(KeyCaps600Creator, ChrID, eventP->data.keyDown.chr);
	    FtrSet(KeyCaps600Creator, ModID, eventP->data.keyDown.modifiers);
	    FtrSet(KeyCaps600Creator, CountID, 1);
	    FtrSet(KeyCaps600Creator, TicksID, TimGetTicks());

	    if( ByteSwap16(0) != eventP->data.keyDown.keyCode )
              FtrSet(KeyCaps600Creator, SuppressHoldID, false);
	  }
	}
  }
  return 0;
}

static Int32 GetTextFieldLen()
{
	UInt16 FormID;
	FieldType *fld;

	FormID = FrmGetActiveFormID();
	if( FormID == 0 )	// Java running, probably
		return -1;

	fld = FrmGetActiveField(NULL);
	if( fld == NULL )
		return -1;	// No length

	return FldGetTextLength(fld);
}

static Boolean isTreo600(KeyCaps600Prefs *prefs)
{
	UInt32 deviceID = 0;
	UInt32 companyID = 0;

	if( prefs->prefsMask & Treo600Override )
		return true;

	if( prefs->prefsMask & Treo650Override )
		return false;

	FtrGet(sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID);
	FtrGet(sysFtrCreator, sysFtrNumOEMCompanyID, &companyID);

	if( (companyID == 'hspr') && (deviceID == 'H101') ) // T600
		return true;

	return false; // future devices are more likely to be 650-like
}

static Int8 SelectPriority(KeyCaps600Prefs *prefs)
{
	int prio = 0;

	if( prefs->prefsMask & LowPrioOverride )
		prio += 32;

	if( prefs->prefsMask & HighPrioOverride )
		prio -= 32;

	return prio;
}

static void butlerNotify(SysNotifyParamType *np)
{
	UInt32 type = np->notifyType;

        np->notifyType = KeyCaps600Creator;
        SysNotifyBroadcast(np);
        np->notifyType = type;
}

#define CREATOR			KeyCaps600Creator

#define STARTUP_CODE_SIZE	152		// size of prc-tools crt0.c
#define FTR_PRIO		998
#define FTR_PTR			999

static void RegisterForNotifications(Int8 priority)
{
	MemHandle hndl;
	void *memptr;
	void *ftrptr;
	void *hn_ptr;

	UInt16 cardNo;
	LocalID dbID;

        DmSearchStateType sstate;

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	// Unregister for notification
	DeregisterNotifications();

	hndl = DmGet1Resource('code',1);
	memptr = MemHandleLock(hndl);

	FtrPtrNew(CREATOR, FTR_PTR, MemHandleSize(hndl),&ftrptr);
	DmWrite(ftrptr, 0, memptr, MemHandleSize(hndl));

	MemHandleUnlock(hndl);

#ifdef PROTECT_DB
	DmDatabaseProtect(cardNo, dbID, true);
#endif

	hn_ptr = ((void *)ftrptr)+STARTUP_CODE_SIZE;

#ifdef DEBUG
	SysNotifyRegister(cardNo, dbID, sysNotifyEventDequeuedEvent, NULL, 
			priority, NULL);
#else
	SysNotifyRegister(cardNo, dbID, sysNotifyEventDequeuedEvent, hn_ptr, 
			priority, NULL);
#endif

	// Set priority feature (we'll use it later to deregister)
	FtrSet(CREATOR, FTR_PRIO, priority);

#ifdef PROTECT_DB
 	SysNotifyRegister(cardNo, dbID, sysNotifyDeleteProtectedEvent, NULL,
			sysNotifyNormalPriority, NULL);
#endif

	// Identify the Phone application
        DmGetNextDatabaseByTypeCreator(true,&sstate,'appl','HsPh',true,&cardNo, &dbID);
	FtrSet(KeyCaps600Creator,CardIdID,cardNo);
	FtrSet(KeyCaps600Creator,DBIdID,dbID);
}

static void DeregisterNotifications()
{	
	Int32 val;
	void *ftrptr;

	UInt16 cardNo;
	LocalID dbID;

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	// Deregister for notification
	val = -100000; FtrGet(CREATOR, FTR_PRIO, &val);

	if( (val >= -127) && (val <= 128) ) // If registered
		SysNotifyUnregister(cardNo, dbID, sysNotifyEventDequeuedEvent, val);

#ifdef PROTECT_DB
 	SysNotifyUnregister(cardNo, dbID, sysNotifyDeleteProtectedEvent,
			sysNotifyNormalPriority);
#endif

	val = 0; FtrGet(CREATOR, FTR_PTR, &val); ftrptr = (void *)val;

	// release feature memory
	if( ftrptr ) 
		FtrPtrFree(CREATOR, FTR_PTR);

#ifdef PROTECT_DB
	DmDatabaseProtect(cardNo, dbID, false);
#endif

	// Flag Butler that we're not here
	FtrSet(KeyCaps600Creator, ButlerFlagID, (DWord)0);
}

static void InitializePrefs(KeyCaps600Prefs *prefs)
{
    Word prefsSize;
    prefsSize = sizeof( KeyCaps600Prefs );

    MemSet(prefs, prefsSize, 0);
    if ((PrefGetAppPreferences(KeyCaps600Creator, 1, prefs, &prefsSize, true)
	 == noPreferenceFound) || (prefsSize < sizeof(prefs))) {
      /* if no prefs, set defaults */
      MemSet(prefs, sizeof(KeyCaps600Prefs), 0);
      prefs->prefsMask = DoubleClickOption | HoldDownCapital | 
	      			DisableInPhone | DetectAutocomplete;
      prefs->timeout = 30;
      prefs->delay   = 30;
      prefs->scrollrepeat = 10;
      prefs->scrolldelay  = 30;
      prefsSize = sizeof( KeyCaps600Prefs );
      /* set prefs so we don't have to do this every time */
      PrefSetAppPreferences(KeyCaps600Creator, 1, 0, prefs, prefsSize, true);
      /* reset all Ftr fields */
      FtrSet(KeyCaps600Creator, ChrID, -1);
      FtrSet(KeyCaps600Creator, CountID, (DWord)0);
      FtrSet(KeyCaps600Creator, TicksID, (DWord)0);
    }
}


static Boolean AboutFormHandleEvent(EventPtr e)
{
  FormPtr frm;
  Boolean handled = false;
  
  switch (e->eType) {
  case frmOpenEvent:
    frm = FrmGetFormPtr(AboutForm);
    FrmDrawForm(frm);
    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case AboutOKButton:
      FrmReturnToForm(0);
      handled = true;
      break;
    }
    break;

  default:
    break;
  }
  return handled;
}

static Boolean AdvancedFormHandleEvent(EventPtr e)
{
  FormPtr frm;
  Boolean handled = false;

  KeyCaps600Prefs prefs;
  Word prefsSize = sizeof(prefs);

  int listopt;

  ControlType *priorityTrigger, *modelTrigger;
  ListPtr priorityList, modelList;
  CharPtr priorityText, modelText;

  InitializePrefs( &prefs );
  //PrefGetAppPreferences(KeyCapsHackCreator, 1, &prefs, &prefsSize, true);
  
  switch (e->eType) {
  case frmOpenEvent:
    frm = FrmGetFormPtr(AdvancedForm);
    FrmSetControlValue(frm, FrmGetObjectIndex(frm, AdvPhoneDisableBox), (prefs.prefsMask & DisableInPhone) ? 1 : 0);
    FrmSetControlValue(frm, FrmGetObjectIndex(frm, AdvDetectAutocompleteBox), (prefs.prefsMask & DetectAutocomplete) ? 1 : 0);

    priorityTrigger = (ControlType*)FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,AdvPriorityTrigger));
    priorityList = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,AdvPriorityList));

    // this doesn't need checking
    listopt = 1 + !!(prefs.prefsMask & LowPrioOverride) - !!(prefs.prefsMask & HighPrioOverride);

    priorityText = LstGetSelectionText(priorityList, listopt);
    CtlSetLabel( priorityTrigger, priorityText );
    LstSetSelection( priorityList, listopt );
    LstMakeItemVisible( priorityList, listopt );

    modelTrigger = (ControlType*)FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,AdvModelTrigger));
    modelList = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,AdvModelList));

    listopt = !!(prefs.prefsMask & Treo600Override) + 2*!!(prefs.prefsMask & Treo650Override);

    // Check invalid
    if( listopt > 2 ) 
    {	
	    // Reset pref
	    prefs.prefsMask &= ~(Treo600Override | Treo650Override);
    	    PrefSetAppPreferences(KeyCapsHackCreator, 1, 0, &prefs, prefsSize, true);
	
	    // Use a valid label
	    listopt = 0;
    }

    modelText = LstGetSelectionText(modelList, listopt);
    CtlSetLabel( modelTrigger, modelText );
    LstSetSelection( modelList, listopt);
    LstMakeItemVisible( modelList, listopt );

    FrmDrawForm(frm);
    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case AdvOKButton:
        FrmReturnToForm(0);
        handled = true;
        break;

    case AdvPhoneDisableBox:
	frm = FrmGetActiveForm();
	prefs.prefsMask &= ~DisableInPhone;
	if (FrmGetControlValue(frm, FrmGetObjectIndex(frm, AdvPhoneDisableBox))) {
	  prefs.prefsMask |= DisableInPhone;
	}
	PrefSetAppPreferences(KeyCapsHackCreator, 1, 0, &prefs, prefsSize, true);
	handled = true;
	break;

    case AdvDetectAutocompleteBox:
	frm = FrmGetActiveForm();
	prefs.prefsMask &= ~DetectAutocomplete;
	if (FrmGetControlValue(frm, FrmGetObjectIndex(frm, AdvDetectAutocompleteBox))) {
	  prefs.prefsMask |= DetectAutocomplete;
	}
	PrefSetAppPreferences(KeyCapsHackCreator, 1, 0, &prefs, prefsSize, true);
	handled = true;
	break;
      } 
  break;

  case popSelectEvent:
    frm = FrmGetActiveForm();
    switch(e->data.popSelect.listID) {
      case AdvPriorityList:
        priorityList = (ListPtr)FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,AdvPriorityList));
	LstSetSelection(priorityList, e->data.popSelect.selection);
	priorityTrigger = (ControlType*)FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,AdvPriorityTrigger));
	CtlSetLabel(priorityTrigger,LstGetSelectionText(priorityList,e->data.popSelect.selection));

	switch(e->data.popSelect.selection)
	{
	  case 0:
	    prefs.prefsMask &= ~LowPrioOverride;
	    prefs.prefsMask |= HighPrioOverride;
	    break;

	  case 1:
	    prefs.prefsMask &= ~LowPrioOverride;
	    prefs.prefsMask &= ~HighPrioOverride;
	    break;

    	  case 2:
	    prefs.prefsMask |= LowPrioOverride;
	    prefs.prefsMask &= ~HighPrioOverride;
	    break;
	}

	if( prefs.prefsMask & KeyCapsEnabled )
		RegisterForNotifications(SelectPriority(&prefs));

	break;

      case AdvModelList:
        modelList = (ListPtr)FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,AdvModelList));
	LstSetSelection(modelList, e->data.popSelect.selection);
	modelTrigger = (ControlType*)FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,AdvModelTrigger));
	CtlSetLabel(modelTrigger,LstGetSelectionText(modelList,e->data.popSelect.selection));

	switch(e->data.popSelect.selection)
	{
	  case 0:
	    prefs.prefsMask &= ~Treo600Override;
	    prefs.prefsMask &= ~Treo650Override;
	    break;

	  case 1:
	    prefs.prefsMask |= Treo600Override;
	    prefs.prefsMask &= ~Treo650Override;
	    break;

	  case 2:
	    prefs.prefsMask &= ~Treo600Override;
	    prefs.prefsMask |= Treo650Override;
	    break;
	}
	break;
    }
    PrefSetAppPreferences(KeyCapsHackCreator, 1, 0, &prefs, prefsSize, true);
    break;


  default:
    break;
  }
  return handled;
}

static Boolean MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;

    FormPtr frm;
    UInt16 cardNo;
    LocalID dbID;
    KeyCaps600Prefs prefs;
    Word prefsSize = sizeof(prefs);
    ControlType *clickTrigger, *holdTrigger, *timeoutTrigger, *delayTrigger, 
			*scrollRepeatTrigger, *scrollDelayTrigger;
    ListPtr clickList, holdList, timeoutList, delayList, scrollRepeatList, scrollDelayList;
    CharPtr triggerText, timeoutText, delayText, 
			scrollDelayText, scrollRepeatText;
    int scrollPeriodIndex;

    InitializePrefs( &prefs );

    // We might have been disabled by an upgrade
    if( prefs.prefsMask & KeyCapsEnabled )
      RegisterForNotifications(SelectPriority(&prefs));

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmSetNavState(frm,kFrmNavStateFlagsObjectFocusMode);

	if (prefs.prefsMask & KeyCapsEnabled) {
	  FrmSetControlValue(frm, FrmGetObjectIndex(frm,MainFormButtonStop),0);
	  FrmSetControlValue(frm, FrmGetObjectIndex(frm,MainFormButtonStart),1);
	} else {
	  FrmSetControlValue(frm, FrmGetObjectIndex(frm,MainFormButtonStart),0);
	  FrmSetControlValue(frm, FrmGetObjectIndex(frm,MainFormButtonStop),1);
	}
	
	/* set trigger labels */
	clickTrigger   = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DoubleClickTrigger));
	clickList      = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DoubleClickList));
	holdTrigger    = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, HoldDownTrigger));
	holdList       = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, HoldDownList));
	timeoutTrigger = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, TimeoutTrigger));
	timeoutList    = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, TimeoutList));
	delayTrigger   = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DelayTrigger));
	delayList      = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DelayList));
	scrollRepeatTrigger = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollRepeatTrigger));
	scrollRepeatList = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollRepeatList));
	scrollDelayTrigger = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollDelayTrigger));
	scrollDelayList = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollDelayList));
	
	timeoutText = LstGetSelectionText(timeoutList, (prefs.timeout)-10);
	CtlSetLabel(timeoutTrigger, timeoutText);
	LstSetSelection(timeoutList, (prefs.timeout)-10);
	LstMakeItemVisible(timeoutList, (prefs.timeout)-10);
	
	delayText = LstGetSelectionText(delayList, (prefs.delay)-10);
	CtlSetLabel(delayTrigger, delayText);
        LstSetSelection(delayList, (prefs.delay-10));
	LstMakeItemVisible(delayList, (prefs.delay)-10);

	switch(prefs.scrollrepeat)
	{
		case 2  : scrollPeriodIndex = 0; break;
		case 3  : scrollPeriodIndex = 1; break;
		case 4  : scrollPeriodIndex = 2; break;
		case 5  : scrollPeriodIndex = 3; break;
		case 7  : scrollPeriodIndex = 4; break;
		case 10 : scrollPeriodIndex = 5; break;
		case 14 : scrollPeriodIndex = 6; break;
		case 20 : scrollPeriodIndex = 7; break;
		case 33 : scrollPeriodIndex = 8; break;
		case 50 : scrollPeriodIndex = 9; break;
		case 100: scrollPeriodIndex = 10; break;
		default:  scrollPeriodIndex = 5; prefs.scrollrepeat = 10; break;
	}
	scrollRepeatText = LstGetSelectionText(scrollRepeatList, scrollPeriodIndex);
	CtlSetLabel(scrollRepeatTrigger, scrollRepeatText);
	LstSetSelection(scrollRepeatList, scrollPeriodIndex);
	LstMakeItemVisible(scrollRepeatList, scrollPeriodIndex);
	
	scrollDelayText = LstGetSelectionText(scrollDelayList, (prefs.scrolldelay)-10);
	CtlSetLabel(scrollDelayTrigger, scrollDelayText);
        LstSetSelection(scrollDelayList, (prefs.scrolldelay-10));
	LstMakeItemVisible(scrollDelayList, (prefs.scrolldelay)-10);

	if (prefs.prefsMask & DoubleClickOption) {
	  if (prefs.prefsMask & DoubleClickRepeat) {
	    LstSetSelection(clickList, 4);
	    triggerText = LstGetSelectionText(clickList, 4);
	    CtlSetLabel(clickTrigger, triggerText);
	  } else {
	    LstSetSelection(clickList, 2);
	    triggerText = LstGetSelectionText(clickList, 2);
	    CtlSetLabel(clickTrigger, triggerText);
	  } 
	} else if (prefs.prefsMask & DoubleClickCapital) {
	  if (prefs.prefsMask & DoubleClickRepeat) {
	    LstSetSelection(clickList, 3);
	    triggerText = LstGetSelectionText(clickList, 3);
	    CtlSetLabel(clickTrigger, triggerText);
	  } else {
	    LstSetSelection(clickList, 1);
	    triggerText = LstGetSelectionText(clickList, 1);
	    CtlSetLabel(clickTrigger, triggerText);
	  } 
	} else {
	  LstSetSelection(clickList, 0);
	  triggerText = LstGetSelectionText(clickList, 0);
	  CtlSetLabel(clickTrigger, triggerText);
	}
	if (prefs.prefsMask & HoldDownOption) {
	  if (prefs.prefsMask & HoldDownRepeat) {
	    LstSetSelection(holdList, 4);
	    triggerText = LstGetSelectionText(holdList, 4);
	    CtlSetLabel(holdTrigger, triggerText);
	  } else {
	    LstSetSelection(holdList, 2);
	    triggerText = LstGetSelectionText(holdList, 2);
	    CtlSetLabel(holdTrigger, triggerText);
	  }
	} else if (prefs.prefsMask & HoldDownCapital) {
	  if (prefs.prefsMask & HoldDownRepeat) {
	    LstSetSelection(holdList, 3);
	    triggerText = LstGetSelectionText(holdList, 3);
	    CtlSetLabel(holdTrigger, triggerText);
	  } else {
	    LstSetSelection(holdList, 1);
	    triggerText = LstGetSelectionText(holdList, 1);
	    CtlSetLabel(holdTrigger, triggerText);
	  }
	} else {
	  LstSetSelection(holdList, 0);
	  triggerText = LstGetSelectionText(holdList, 0);
	  CtlSetLabel(holdTrigger, triggerText);
	}

	FrmSetControlValue(frm, FrmGetObjectIndex(frm, ResetEnableBox), (prefs.prefsMask & EnableOnReset) ? 1 : 0);

	FrmDrawForm(frm);	
	handled = true;
	break;
	
    case menuEvent:
      MenuEraseStatus(NULL);
      switch(e->data.menu.itemID) {
      case MenuAbout:
	FrmPopupForm(AboutForm);
	handled = true;
	break;
      case MenuAdvanced:
	FrmPopupForm(AdvancedForm);
	handled = true;
	break;
      }
      
      
    case ctlSelectEvent:
      switch(e->data.ctlSelect.controlID) {
	
      case MainFormButtonStart:
	/* register for notifications */
	if (0 != SysCurAppDatabase(&cardNo, &dbID))
	  return -1;  

	
	RegisterForNotifications(SelectPriority(&prefs));
	// Flag Butler that we're here
	FtrSet(KeyCaps600Creator, ButlerFlagID, !!( prefs.prefsMask & HoldDownMask ));

	prefs.prefsMask |= KeyCapsEnabled;
	PrefSetAppPreferences(KeyCaps600Creator, 1, 0, &prefs, prefsSize, true);
	frm = FrmGetActiveForm();
	FrmSetControlValue(frm, FrmGetObjectIndex(frm,MainFormButtonStop),0);
	break;
	
      case MainFormButtonStop:
	/* unregister */
	if (0 != SysCurAppDatabase(&cardNo, &dbID))
	  return -1;  

        DeregisterNotifications();
	prefs.prefsMask &= ~KeyCapsEnabled;
	PrefSetAppPreferences(KeyCaps600Creator, 1, 0, &prefs, prefsSize, true);
	frm = FrmGetActiveForm();
	FrmSetControlValue(frm, FrmGetObjectIndex(frm,MainFormButtonStart),0);
	break;

      case ResetEnableBox:
	frm = FrmGetActiveForm();
	prefs.prefsMask &= ~EnableOnReset;
	if (FrmGetControlValue(frm, FrmGetObjectIndex(frm, ResetEnableBox))) {
	  prefs.prefsMask |= EnableOnReset;
	}
	PrefSetAppPreferences(KeyCapsHackCreator, 1, 0, &prefs, prefsSize, true);
	handled = true;
	break;
    }

      break;
      
    case popSelectEvent:
      switch(e->data.popSelect.listID) {
      case DoubleClickList:
	frm = FrmGetActiveForm();
	clickList    = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DoubleClickList));
	LstSetSelection(clickList, e->data.popSelect.selection);
	clickTrigger   = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DoubleClickTrigger));
	CtlSetLabel(clickTrigger, LstGetSelectionText(clickList, e->data.popSelect.selection));
	prefs.prefsMask &= ~DoubleClickMask;
	switch(e->data.popSelect.selection)
	  {
	  case 4:
	    prefs.prefsMask |= DoubleClickOption;
	    prefs.prefsMask |= DoubleClickRepeat;
	    break;
	  case 3:
	    prefs.prefsMask |= DoubleClickCapital;
	    prefs.prefsMask |= DoubleClickRepeat;
	    break;
	  case 2:
	    prefs.prefsMask |= DoubleClickOption;
	    prefs.prefsMask &= ~DoubleClickRepeat;
	    break;
	  case 1:
	    prefs.prefsMask |= DoubleClickCapital;
	    prefs.prefsMask &= ~DoubleClickRepeat;
	    break;
	  default:
	    break;
	  }
	break;
      case HoldDownList:
	frm = FrmGetActiveForm();
	holdList     = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, HoldDownList));
	LstSetSelection(holdList, e->data.popSelect.selection);
	holdTrigger   = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, HoldDownTrigger));
	CtlSetLabel(holdTrigger, LstGetSelectionText(holdList, e->data.popSelect.selection));
	prefs.prefsMask &= ~HoldDownMask;
	switch(e->data.popSelect.selection)
	  {
	  case 4:
	    prefs.prefsMask |= HoldDownOption;
	    prefs.prefsMask |= HoldDownRepeat;
	    break;
	  case 3:
	    prefs.prefsMask |= HoldDownCapital;
	    prefs.prefsMask |= HoldDownRepeat;
	    break;
	  case 2:
	    prefs.prefsMask |= HoldDownOption;
	    prefs.prefsMask &= ~HoldDownRepeat;
	    break;
	  case 1:
	    prefs.prefsMask |= HoldDownCapital;
	    prefs.prefsMask &= ~HoldDownRepeat;
	    break;
	  default:
	    break;
	  }

	// Flag Butler that we're here
	FtrSet(KeyCaps600Creator, ButlerFlagID, !!( prefs.prefsMask & HoldDownMask ));

	break;
      case TimeoutList:
	frm = FrmGetActiveForm();
	timeoutList  = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, TimeoutList));
	LstSetSelection(timeoutList, e->data.popSelect.selection);
	timeoutTrigger   = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, TimeoutTrigger));
	CtlSetLabel(timeoutTrigger, LstGetSelectionText(timeoutList, e->data.popSelect.selection));
	prefs.timeout = (e->data.popSelect.selection + 10);
	break;
      case DelayList:
	frm = FrmGetActiveForm();
	delayList  = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DelayList));
	LstSetSelection(delayList, e->data.popSelect.selection);
	delayTrigger   = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, DelayTrigger));
	CtlSetLabel(delayTrigger, LstGetSelectionText(delayList, e->data.popSelect.selection));
	prefs.delay = (e->data.popSelect.selection + 10);
	break; 
      case ScrollRepeatList:
	frm = FrmGetActiveForm();
	scrollRepeatList = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollRepeatList));
	LstSetSelection(scrollRepeatList, e->data.popSelect.selection);
	scrollRepeatTrigger = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollRepeatTrigger));
	CtlSetLabel(scrollRepeatTrigger, 
		LstGetSelectionText(scrollRepeatList, e->data.popSelect.selection));
	switch( e->data.popSelect.selection )
	{
		case 0:	prefs.scrollrepeat = 2; break;
		case 1:	prefs.scrollrepeat = 3; break;
		case 2: prefs.scrollrepeat = 4; break;
		case 3: prefs.scrollrepeat = 5; break;
		case 4: prefs.scrollrepeat = 7; break;
		case 5: prefs.scrollrepeat = 10; break;
		case 6: prefs.scrollrepeat = 14; break;
		case 7: prefs.scrollrepeat = 20; break;
		case 8: prefs.scrollrepeat = 33; break;
		case 9: prefs.scrollrepeat = 50; break;
		case 10: prefs.scrollrepeat = 100; break;
	}
        break;
      case ScrollDelayList:
        frm = FrmGetActiveForm();
	scrollDelayList = (ListPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollDelayList));
	LstSetSelection(scrollDelayList, e->data.popSelect.selection);
	scrollDelayTrigger = (ControlType*)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ScrollDelayTrigger));
	CtlSetLabel(scrollDelayTrigger, 
		LstGetSelectionText(scrollDelayList, e->data.popSelect.selection));
	prefs.scrolldelay = (e->data.popSelect.selection + 10);
        break;
      }
      PrefSetAppPreferences(KeyCapsHackCreator, 1, 0, &prefs, prefsSize, true);
      FtrSet(KeyCaps600Creator, RepeatSetID, false);	// force repeat rate to be set
      handled = true;
      break;
    default:
      break;
    }
    
    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case MainForm:
	  FrmSetEventHandler(frm, MainFormHandleEvent);
	  break;
	case AdvancedForm:
	  FrmSetEventHandler(frm, AdvancedFormHandleEvent);
	  break;
	case AboutForm:
	  FrmSetEventHandler(frm, AboutFormHandleEvent);
	  break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;
    KeyCaps600Prefs prefs;
    Word prefsSize = sizeof(prefs);
    LocalID dbID;
    UInt16 cardNo;

    switch (cmd) {
    case sysAppLaunchCmdNormalLaunch:
      err = StartApplication();
      if (err) return err;
      
      EventLoop();
      StopApplication();
      break;
      
    case sysAppLaunchCmdNotify:
      HandleNotification(cmdPBP);
      break;

    case sysAppLaunchCmdSystemReset:
      /* get the prefs */
      InitializePrefs( &prefs ) ;

      // If we're not supposed to enable on reset, mark as disabled
      if( ! (prefs.prefsMask & EnableOnReset) )
      {
        prefs.prefsMask &= ~KeyCapsEnabled;
        PrefSetAppPreferences(KeyCaps600Creator, 1, 0, &prefs, prefsSize, true);
        break; 
      }

      if( prefs.prefsMask & KeyCapsEnabled ) {
	if (0 != SysCurAppDatabase(&cardNo, &dbID))
	  return -1;  
        RegisterForNotifications(SelectPriority(&prefs));
	// Flag Butler that we're here
	FtrSet(KeyCaps600Creator, ButlerFlagID, !!( prefs.prefsMask & HoldDownMask ));

	prefs.prefsMask |= KeyCapsEnabled;
	PrefSetAppPreferences(KeyCaps600Creator, 1, 0, &prefs, prefsSize, true);
      } 
      break;

    default:
      return sysErrParamErr;
      break;
    }

    return 0;
}
